/**
 * @file
 *
 * OpenCL example
 *
 * (c) 2010 Jens Teubner, ETH Zurich, Systems Group
 *
 * $Id: coalescing.c 3188 2017-07-17 11:51:21Z jens.teubner $
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>

#include <mach/mach_time.h>
#include <time.h>

#include <OpenCL/opencl.h>

/**
 * Logging convenience macro (can turn off logging by re-defining this macro).
 */
#define LOG(...)  do {             \
    fprintf (stderr, __VA_ARGS__); \
    fflush (stderr);               \
  } while (0)

#define DEBUG 0
#define USE_OPENCL 1

#define KERNEL_SOURCE "coalescing.cl"
#define NITEMS (1 << 24)

static unsigned int *
generate_data (unsigned int nitems)
{
    unsigned int  *ret;

    ret = malloc (nitems * sizeof (*ret));
    if (! ret)
    {
        fprintf (stderr, "error allocating memory\n");
        exit (EXIT_FAILURE);
    }
    for (unsigned int i = 0; i < nitems; i++)
    {
        ret[i] = random () % 10;
    }

    return ret;
}

static void
mach_absolute_difference (uint64_t end, uint64_t start, struct timespec *tp)
{
    uint64_t                          difference = end - start;  
    static mach_timebase_info_data_t  info       = {0,0};  

    if (info.denom == 0)
        mach_timebase_info (&info);  

    uint64_t elapsednano = difference * (info.numer / info.denom);  

    tp->tv_sec  = elapsednano * 1e-9;  
    tp->tv_nsec = elapsednano - (tp->tv_sec * 1e9);  
}  

static cl_device_id
query_devices (void)
{
    cl_device_id  ret;

    cl_int        err;                /**< error code from OpenCL calls */
    cl_uint       num_devices;        /**< # of devices in the system */
    char          device_name[1024];  /**< device name */

    /* have a variable ready for the different result types needed */
    size_t        returned_size;
    cl_ulong      ret_ulong;
    cl_uint       ret_uint;
    size_t        ret_sizet;

    /*
     * Let's see what devices there are in the system
     *
     * (This call is all we really need in this function!)
     */
    err = clGetDeviceIDs (NULL,               /* platform id (or NULL) */
                          CL_DEVICE_TYPE_GPU, /* device type */
                          1,                  /* number of devices to report */
                          &ret,               /* list of OpenCL devices found */
                          &num_devices);      /* number of devices found */
    if (err != CL_SUCCESS)
    {
        fprintf (stderr, "error in clGetDeviceIDs!\n");
        exit (EXIT_FAILURE);
    }

    LOG("Found %i device(s) in your system.\n", num_devices);

    if (num_devices == 0)
    {
        fprintf (stderr, "No usable device found.\n");
        exit (EXIT_FAILURE);
    }

    /*
     * Report information about the device we found.
     */
    err = clGetDeviceInfo (ret,              /* device to query */
                           CL_DEVICE_NAME,   /* type of information to query */
                           sizeof (device_name),   /* size of return buffer */
                           device_name,      /* return buffer */
                           &returned_size);  /* bytes written into return buf */
    if (err != CL_SUCCESS)
    {
        fprintf (stderr, "error in clGetDeviceInfo!\n");
        exit (EXIT_FAILURE);
    }

    /* zero terminate the string (play safe) */
    device_name[returned_size] = '\0';

    LOG("Your device is a %s.\n", device_name);

    err = clGetDeviceInfo (ret,
                           CL_DEVICE_GLOBAL_MEM_SIZE,
                           sizeof (ret_ulong),
                           &ret_ulong,
                           &returned_size);
    if (err != CL_SUCCESS)
    {
        fprintf (stderr, "error in clGetDeviceInfo!\n");
        exit (EXIT_FAILURE);
    }

    LOG("Your device hosts %llu bytes of global memory.\n", ret_ulong);

    err = clGetDeviceInfo (ret,
                           CL_DEVICE_MAX_COMPUTE_UNITS,
                           sizeof (ret_uint),
                           &ret_uint,
                           &returned_size);
    if (err != CL_SUCCESS)
    {
        fprintf (stderr, "error in clGetDeviceInfo!\n");
        exit (EXIT_FAILURE);
    }

    LOG("You can run at most %i parallel compute units.\n", ret_uint);

    err = clGetDeviceInfo (ret,
                           CL_DEVICE_MAX_WORK_GROUP_SIZE,
                           sizeof (ret_sizet),
                           &ret_sizet,
                           &returned_size);
    if (err != CL_SUCCESS)
    {
        fprintf (stderr, "error in clGetDeviceInfo!\n");
        exit (EXIT_FAILURE);
    }

    LOG("There can be at most %lu items in a work-group.\n", ret_sizet);

    return ret;
}

static char *
load_program_from_source_file (const char *filename)
{
    struct stat  statbuf;
    FILE        *f;
    char        *ret;

    f = fopen (filename, "r");
    if (!f)
    {
        fprintf (stderr, "Cannot open %s.\n", filename);
        exit (EXIT_FAILURE);
    }

    stat (filename, &statbuf);

    ret = (char *) malloc (statbuf.st_size + 1);

    fread (ret, statbuf.st_size, 1, f);

    ret[statbuf.st_size] = '\0';

    fclose (f);

    return ret;
}

int
main (void)
{
    cl_int err;                       /**< error code from OpenCL calls */

    unsigned int       *items;              /**< actual data to process */
    const unsigned int  nitems = NITEMS;

    cl_device_id  device_id;          /**< Device ID */

    cl_context    context;            /**< compute context */
    char         *prog_src;           /**< compute kernel source */
    cl_program    program;            /**< compute program (whole program) */
    cl_kernel     kernel;             /**< compute kernel (single func) */

    cl_mem        din;                /**< device memory region (input) */
    cl_mem        dout;               /**< device memory region (output) */

    cl_command_queue commandQueue;    /**< command queue for device */

    size_t        global;             /**< global number of work items */

    uint64_t      t_start, t_end;
    struct timespec time_diff;

    /*
     * Allocate host memory for input data and populate array with
     * random data.
     */
    items = generate_data (nitems);

    /*
     * See what devices we have out there.
     */
    device_id = query_devices ();

    /*
     * Create a compute context.
     */
    context = clCreateContext (NULL,        /* choose any platform */
                               1,           /* number of devices */
                               &device_id,  /* list of devices */
                               NULL,        /* callback function that can be
                                               registered by the application */
                               NULL,        /* user data that would be passed
                                               to the callback function */
                               &err);       /* error return code */
    if (err != CL_SUCCESS)
    {
        fprintf (stderr, "error in clCreateContext!\n");
        exit (EXIT_FAILURE);
    }

    LOG("Compute context created.\n");

    /*
     * Load the kernel source from a file.
     */
    prog_src = load_program_from_source_file (KERNEL_SOURCE);

    /*
     * Create the compute program from the source buffer
     */
    program = clCreateProgramWithSource (
            context,             /* a valid compute context */
            1,                   /* number zero-term. strings in prog_src */
            (const char **) &prog_src,   /* (array of) program source strings */
            NULL,                /* lengths of all strings passed */
            &err);               /* error code */
    if (err != CL_SUCCESS)
    {
        fprintf (stderr, "error in clCreateProgramWithSource!\n");
        exit (EXIT_FAILURE);
    }

    LOG("Compute program created.\n");

    /*
     * Compile/build the program.
     */
    err = clBuildProgram (program,
                          0,        /* number of devices in device list */
                          NULL,     /* device list; NULL -> build for all dev.
                                       of the context */
                          NULL,     /* build options */
                          NULL,     /* callback for notification routine */
                          NULL);    /* user data for notification routine */
    if (err != CL_SUCCESS)
    {
        size_t len;
        char   buf[2048];

        fprintf (stderr, "error in clCreateProgramWithSource!\n");

        /* Print build errors (helps debugging when something went wrong). */
        clGetProgramBuildInfo (program, device_id, CL_PROGRAM_BUILD_LOG,
                               sizeof (buf), buf, &len);
        buf[sizeof(buf)-1] = '\0';   /* play safe */

        fprintf (stderr, "%s\n", buf);

        exit (EXIT_FAILURE);
    }

    LOG("Program compiled.\n");

    /*
     * Create the compute kernel that we wish to run.
     */
    kernel = clCreateKernel (program, "copy", &err);
    if (err != CL_SUCCESS)
    {
        fprintf (stderr, "error in clCreateKernel!\n");
        exit (EXIT_FAILURE);
    }

    LOG("Kernel created.\n");

    /*
     * Create a memory buffers.
     */
    din = clCreateBuffer (context, CL_MEM_READ_WRITE,
                          nitems * sizeof (*items),
                          NULL,                         /* host pointer */
                          &err);
    if (err != CL_SUCCESS)
    {
        fprintf (stderr, "error in clCreateBuffer!\n");
        exit (EXIT_FAILURE);
    }

    LOG("Device memory region allocated (input).\n");

    dout = clCreateBuffer (context, CL_MEM_READ_WRITE,
                           (nitems + 50) * sizeof (*items),
                           NULL,                         /* host pointer */
                           &err);
    if (err != CL_SUCCESS)
    {
        fprintf (stderr, "error in clCreateBuffer!\n");
        exit (EXIT_FAILURE);
    }

    LOG("Device memory region allocated (output).\n");

    /*
     * Create command queue.
     */
    commandQueue = clCreateCommandQueue (context, device_id,
                                         0,  /* properties */
                                         &err);
    if (err != CL_SUCCESS)
    {
        fprintf (stderr, "error in clCreateCommandQueue!\n");
        exit (EXIT_FAILURE);
    }
    LOG("Command queue created.\n");


    /*
     * Fill device memory with data.
     */
    err = clEnqueueWriteBuffer (
            commandQueue,
            din,
            CL_TRUE,                    /* blocking write */
            0,                          /* offset */
            nitems * sizeof (*items),   /* data region size */
            items,                      /* source memory region */
            0,                          /* events to wait for before exec. */
            NULL,                       /* event wait list */
            NULL);                      /* event id for this request (can be
                                           used to query status of request */
    if (err != CL_SUCCESS)
    {
        fprintf (stderr, "error in clEnqueueWriteBuffer!\n");
        exit (EXIT_FAILURE);
    }

    LOG("Data copied over to device.\n");

    for (unsigned int offset = 0; offset < 35; offset++)
    {
        LOG("Running with offset %u\n", offset);

        for (unsigned int i = 0; i < 5; i++)
        {
            err = 0;
            err |= clSetKernelArg (kernel, 0, sizeof (cl_mem), &din);
            err |= clSetKernelArg (kernel, 1, sizeof (cl_mem), &dout);
            err |= clSetKernelArg (kernel, 2, sizeof (unsigned int), &offset);
            if (err != CL_SUCCESS)
            {
                fprintf (stderr, "error in clSetKernelArg!\n");
                exit (EXIT_FAILURE);
            }

            t_start = mach_absolute_time ();

            for (unsigned int j = 0; j < 10; j++)
            {
                /*
                 * Fire off a kernel!
                 */
                global = nitems;           /* global is size_t; n is uint */
                err = clEnqueueNDRangeKernel (
                        commandQueue,
                        kernel,
                        1,                /* number of dimensions */
                        NULL,             /* change IDs in kernels? */
                        &global,          /* work size */
                        NULL,             /* size of a work-group; NULL: let the
                                             runtime decide */
                        0, NULL, NULL);   /* don't use any event stuff */
                if (err != CL_SUCCESS)
                {
                    fprintf (stderr, "error in clEnqueueNDRangeKernel!\n");
                    exit (EXIT_FAILURE);
                }
            }

            /* Wait for kernel run to finish */
            clFinish (commandQueue);

            t_end = mach_absolute_time ();

            mach_absolute_difference (t_end, t_start, &time_diff);

            if (i)
                printf ("%u %lu %lu\n",
                        offset, time_diff.tv_sec, time_diff.tv_nsec);
        }

    }

    LOG("Done.");

    /* ----- Now release all the data structures that we allocated. ----- */

    /* Release command queue */
    err = clReleaseCommandQueue (commandQueue);
    if (err != CL_SUCCESS)
    {
        fprintf (stderr, "error in clReleaseCommandQueue!\n");
    }
    LOG("Command queue released.\n");

    /* Release buffer */
    err = clReleaseMemObject (din);
    if (err != CL_SUCCESS)
    {
        fprintf (stderr, "error in clReleaseMemObject!\n");
    }
    LOG("Device memory region released.\n");

    err = clReleaseMemObject (dout);
    if (err != CL_SUCCESS)
    {
        fprintf (stderr, "error in clReleaseMemObject!\n");
    }
    LOG("Device memory region released.\n");

    /* Release kernel */
    err = clReleaseKernel (kernel);
    if (err != CL_SUCCESS)
    {
        fprintf (stderr, "error in clReleaseKernel!\n");
    }
    LOG("Compute kernel released.\n");

    /* Release the compute context.  */
    err = clReleaseProgram (program);
    if (err != CL_SUCCESS)
    {
        fprintf (stderr, "error in clReleaseProgram!\n");
    }
    LOG("Compute program released.\n");

    /* free memory allocated for kernel source code */
    free (prog_src);

    /* Release the compute context.  */
    err = clReleaseContext (context);
    if (err != CL_SUCCESS)
    {
        fprintf (stderr, "error in clReleaseContext!\n");
    }
    LOG("Compute context released.\n");

    LOG("Goodbye.\n");

    return EXIT_SUCCESS;
}
